uplevel #0 {
	global	auto_path env;

	if {[lsearch -exact $auto_path [file dirname [info script]]] < 0} {
		lappend auto_path [file dirname [info script]];
	}

	set {} [file tail [file rootname [info script]]];
	
	while {[file rootname ${}] != ${}} {
		set {} [file rootname ${}];
	}

	package provide ${} 1.0;

	proc ${} {args} "\n\
	if {!\[llength \$args\]} {\n\
		error \"${} option ?arg arg ...?\";\n\
	};\n\
	set option	\[lindex \$args 0\];\n\
	set args		\[lrange \$args 1 end\];\n\
	if {\[::${}::utils::isValidOption \$option\]} {\n\
		set ${}::_level	\[uplevel info level\];\n\
			if {\[set ${}::_level\]} {\n\
				set ${}::_caller	\[uplevel info level \[set ${}::_level\]\];\n\
			} else {\n\
				set ${}::_caller	{};\n\
			}\n\
			set ${}::_namespace	\[uplevel namespace current\];\n\
		return \[eval ::${}::_\[::${}::utils;;getValidOption \$option\] \$args\];\n\
	};\n\
	error \[::${}::utils::badOption \$option \[set ::${}::_subCommands\]\];\n\
";

	namespace eval ${} {
		proc this {} "return [namespace current]";
		proc parent	{} "return [namespace parent [this]]";

		proc isTkPresent {} {return [expr {[info commands ::tk] != {}}];};

		namespace eval utils {
			proc this {} "return [namespace current]";
			proc parent	{} "return [namespace parent [this]]";

			variable _options;

			proc buildOptions {} {
				variable _options;

				set _options	{};

				foreach option [info commands "[parent]::_\[a-z\]*"] {
					lappend _options [string trimleft [namespace tail $option] "_"];
				}

				set _options	[lsort $_options];

				return;
			}

			proc badOption {option {options {}}} {
				if {![llength $options]} {
					variable _options;

					set options	$_options;
				}

				if {[llength $options] == 1} {
					return "bad option \"$option\": must be $options";
				}

				return "bad option \"$option\": must be [join [linsert [join [lsort -dictionary $options] {, }] end-1 {or}]]";
			}

			proc isValidOption {option {options {}}} {
				if {$option == ""} {
					return {};
				}

				if {$options == ""} {
					variable _options;

					set options	$_options;
				}		
				
				if {[lsearch -glob $options ${option}*] >= 0} {
					return 1;
				}

				return 0;
			}

			proc getValidOption {option {options {}}} {
				if {$option == ""} {
					return {};
				}

				if {$options == ""} {
					variable _options;

					set options	$_options;
				}		
				
				if {[set optionsIdx [lsearch -glob $options ${option}*]] >= 0} {
					return [lindex $options $optionsIdx];
				}

				return {};
			}

			proc stdch {channel args} {
				if {$args == ""} {
					puts $channel "";

					return;
				}

				if {[lindex $args 0] == "-nonewline"} {
					set nonewline	1;
					set args			[lrange $args 1 end];
				} else {
					set nonewline	0;
				}

				if {$nonewline} {
					puts -nonewline $channel [join $args "\n"];
				} else {
					puts $channel [join $args "\n"];
				}

				flush $channel;

				return;
			} 

			proc stdout {args} {
				return [eval stdch stdout $args];
			}

			proc stderr {args} {
				return [eval stdch stderr $args];
			}

			namespace export stdch stdout stderr;
			namespace inscope [parent] [list namespace import -force [this]::*];
		} 

		variable _state;
		variable _index;
		variable _level;
		variable _caller;
		variable _namespace;
		variable _handlers;
		variable _registeredHandlers;

		proc _add {args} {
			if {![set argsc [llength $args]] || ($argsc > 2)} {
				error "wrong # args: should be \"[this] add ?priority? command\"";
			}

			if {$argsc == 1} {
				set priority	0;
				set command		[lindex $args 0];
			} else {
				foreach {priority command} $args {break;};
			}

			if {![string is integer -strict $priority] ||
				 ($priority < 0)} {
				error "expected integer value greater or equal than 0, but got \"$priority\"";
			}

			if {($command == "") || ![info complete $command]} {
				error "expected valid command, but got \"$command\"";
			}

			variable _index;
			variable _handlers;

			set token	[format "[this]CB%05d" $_index];

			incr _index;

			set _handlers($token)	[list $priority $command];

			if {![info exists _handlers($priority)]} {
				set _handlers($priority)	{};
			}

			set _handlers($priority)	[linsert $_handlers($priority) -1 $token];

			return $token;
		}

		proc _delete {token} {
			if {[catch {_exists $token;} reason]} {
				error $reason;
			}

			if {!$reason} {
				error "no such token";
			}

			if {[_Isregistered $token]} {
				_unregister $token;
			}

			variable _handlers;

			set priority					[lindex $_handlers($token) 0];
			set _handlers($priority)	[lreplace \
				$_handlers($priority) \
				[set idx [lsearch -exact $_handlers($priority) $token]] $idx \
			];

			array unset _handlers $token;

			return;
		}

		proc _Execute {token {ignoreErrors 0}} {
			variable _level;

			set command	[_PrepareCommand $token];

			if {[catch {uplevel #$_level $command} reason] && !$ignoreErrors} {
				error "couldn't exit, aborting execution of exit callback: $reason";
			}

			return;
		} 

		proc _exists {token} {
			variable _handlers;
			
			if {[array names _handlers $token] == ""} {
				return 0;
			}

			return 1;
		}

		proc _exit {args} {
			if {[set argsc [llength $args]] > 3} {
				error "wrong # args: should be \"exit ?-passby? ?-ignoreerrors? ?returnCode?\"";
			}

			set returnCode		0;
			set passBy			0;
			set ignoreErrors	1;

			set idx	0;

			foreach arg $args {
				if {[string is integer -strict $arg]} {
					set returnCode	$arg;

					incr idx;

					break;
				} elseif {([set optLength [string length $arg]] >= 2) &&
							 [string equal -length $optLength "-passby" $arg]} {
					set passBy	1;
				} elseif {($optLength >= 2) &&
							 [string equal -length $optLength "-ignoreerrors" $arg]} {
					set ignoreErrors	1;
				} else {
					error "bad option \"$arg\": must be -ignoreerrors, or -passby";
				}

				incr idx;
			}

			if {$idx != $argsc} {
				error "wrong # args: should be \"exit ?-passby? ?-ignoreerrors? ?returnCode?\"";
			}

			if {!$passBy} {
				variable _handlers;

				foreach priority [lsort -integer -decreasing [array names _handlers {[0-9]*}]] {
					foreach token $_handlers($priority) {
						if {![_Isregistered $token]} {
							_Execute $token $ignoreErrors;
						}
					}
				}
			}

			_Exit $returnCode;

			return;
		}

		proc _get {args} {
			if {$args == ""} {
				error "wrong # args: should be \"[this] get option ?arg arg ...?";
			}

			set option	[lindex $args 0];
			set options	{handler registered};
			set args		[lrange $args 1 end];

			if {[utils::isValidOption $option $options]} {
				return [eval _Get[utils::getValidOption $option $options] $args];
			}

			return -code error [utils::badOption $option $options];
		}

		proc _Gethandler {args} {
			if {[llength $args] > 2} {
				error "wrong # args: should be \"[this] get handler ?option ?arg arg ...??";
			}

			set option	[lindex $args 0];
			set options	{command count priority tokens};
			set args		[lrange $args 1 end];

			if {[utils::isValidOption $option $options]} {
				return [eval _GetHandler[utils::getValidOption $option $options] $args];
			}

			return -code error [utils::badOption $option $options];
		}

		proc _GetHandlercount {} {
			variable _handlers;

			return [llength [array names _handlers {At*}]];
		}

		proc _GetHandlercommand {token} {
			if {[catch {_exists $token;} reason]} {
				error $reason;
			}

			if {!$reason} {
				error "no such token";
			}

			variable _handlers;

			return [lindex $_handlers($token) 1];
		}

		proc _GetHandlerpriority {token} {
			if {[catch {_exists $token;} reason]} {
				error $reason;
			}

			if {!$reason} {
				error "no such token";
			}

			variable _handlers;

			return [lindex $_handlers($token) 0];
		}

		proc _GetHandlertokens {} {
			variable _handlers;

			return [lsort -dictionary [array names _handlers [this]*]];
		}

		proc _Getregistered {args} {
			if {[llength $args] > 2} {
				error "wrong # args: should be \"[this] get registered ?option ?arg arg ...??";
			}

			set option	[lindex $args 0];
			set options	{count last token tokens widgets};
			set args		[lrange $args 1 end];

			if {[utils::isValidOption $option $options]} {
				return [eval _GetRegistered[utils::getValidOption $option $options] $args];
			}

			return -code error [utils::badOption $option $options];
		}

		proc _GetRegisteredcount {} {
			variable _registeredHandlers;

			return [llength [array names _registeredHandlers {.*}]];
		}

		proc _GetRegisteredlast {widget} {
			if {![_Isregistered $widget]} {
				error "no such registered widget \"$widget\"";
			}

			variable _registeredHandlers;

			return [lindex $_registeredHandlers($widget) 1];
		}

		proc _GetRegisteredtoken {widget} {
			if {![_Isregistered $widget]} {
				error "no such registered widget \"$widget\"";
			}

			variable _registeredHandlers;

			return [lindex $_registeredHandlers($widget) 0];
		}

		proc _GetRegisteredtokens {} {
			variable _registeredHandlers;

			return [lsort -dictionary [array names _registeredHandlers [this]*]];
		}

		proc _GetRegisteredwidget {token} {
			if {![_Isregistered $token]} {
				error "no such registered token \"$token\"";
			}

			variable _registeredHandlers;

		}

		proc _GetRegisteredwidgets {{token {}}} {
			variable _registeredHandlers;

			if {$token != ""} {
				return [lsort -dictionary $_registeredHandlers($token)];
			}

			return [lsort -dictionary [array names _registeredHandlers {.*}]];
		}

		proc _is {args} {
			if {$args == ""} {
				error "wrong # args: should be \"[this] get option ?arg arg ...?";
			}

			set option	[lindex $args 0];
			set options	{off on registered};
			set args		[lrange $args 1 end];

			if {[utils::isValidOption $option $options]} {
				return [eval _Is[utils::getValidOption $option $options] $args];
			}

			return -code error [utils::badOption $option $options];
		}

		proc _Isoff {} {
			variable _state;

			return [expr {$_state == 0}];
		}

		proc _Ison {} {
			variable _state;

			return [expr {$_state != 0}];
		}

		proc _Isregistered {widgetOrToken} {
			variable _registeredHandlers;

			return [expr {[array names _registeredHandlers $widgetOrToken] != ""}];
		}

		proc _off {} {
			variable _state;

			if {!$_state} {
				return 0;
			}

			interp alias {} ::exit {};
			rename [this]::_Exit ::exit;

			return [set _state 0];
		}

		proc _on {} {
			variable _state;

			if {$_state} {
				return 1;
			}

			rename ::exit [this]::_Exit;
			interp alias {} ::exit {} [this]::_exit;

			return [set _state 1];
		}

		proc _PrepareCommand {token} {
			variable _level;
			variable _caller;
			variable _namespace;
			variable _handlers;

			set command		[lindex $_handlers($token) 1];

			regsub -all -- {%p} $command [lindex $_handlers($token) 0] command;
			regsub -all -- {%t} $command $token                               command;
			regsub -all -- {%l} $command $_level                              command;
			regsub -all -- {%c} $command [lindex $_caller 0]                  command;
			regsub -all -- {%a} $command [lrange $_caller 1 end]              command;
			regsub -all -- {%n} $command $_namespace                          command;

			return $command;
		}

		proc _register {widget {token {}}} {
			if {![isTkPresent]} {
				error "package Tk needed to register exit handler \"$token\" to widget \"$widget\"";
			}

			variable _registeredHandlers;

			if {$token == ""} {
				set command	"::[this] exit";
			} else {
				if {[catch {_exists $token;} reason]} {
					error $reason;
				}

				if {!$reason} {
					error "no such token";
				}

				if {[winfo toplevel $widget] != $widget} {
					error "expected toplevel widget, but got widget \"$widget\" with class [winfo class $widget]";
				}

				set command	"::[this]::_Execute $token 1; ::[this] unregister $widget; destroy $widget;";

				lappend _registeredHandlers($token)	$widget;
			}

			set _registeredHandlers($widget) [list $token [wm protocol $widget WM_DELETE_WINDOW]];

			wm protocol $widget WM_DELETE_WINDOW $command;

			return;
		}

		proc _reset {} {
			if {$args == ""} {
				error "wrong # args: should be \"[this] get option ?arg arg ...?";
			}

			set option	[lindex $args 0];
			set options	{all handlers registered};
			set args		[lrange $args 1 end];

			if {[utils::isValidOption $option $options]} {
				return [eval _Reset[utils::getValidOption $option $options] $args];
			}

			return -code error [utils::badOption $option $options];
		}

		proc _Resetall {} {
			_Resetregistered;
			_Resethandlers;

			return;
		}

		proc _Resethandlers {} {
			foreach token [_GetHandlertokens] {
				_delete $token;
			}

			return;
		}

		proc _Resetregistered {} {
			foreach widget [_GetRegisteredwidgets] {
				_unregister $widget;
			}

			return;
		}

		proc _state {{state {}}} {
			variable _state;

			if {$state == ""} {
				return $_state;
			}

			if {![string is boolean $state]} {
				error "expected boolean, but got \"$state\"";
			}

			return [set _state [expr {$state != 0}]];
		}

		proc _unload {} {
			~[this];

			return;
		}

		proc _unregister {widgetOrToken} {
			if {![isTkPresent]} {
				error "package Tk needed to register exit handler \"$token\" to widget \"$widget\"";
			}

			if {![_Isregistered $widgetOrToken]} {
				error "no such registered widget or token \"$widgetOrToken\"";
			}

			variable _registeredHandlers;

			if {[winfo exists $widgetOrToken]} {
				wm protocol $widgetOrToken WM_DELETE_WINDOW [lindex $_registeredHandlers($widgetOrToken) 1];

				if {[set token [lindex $_registeredHandlers($widgetOrToken) 0]] != ""} {
					set _registeredHandlers($token)	[lreplace \
						$_registeredHandlers($token) \
						[set idx [lsearch -exact $_registeredHandlers($token) $widgetOrToken]] $idx \
					];

					if {$_registeredHandlers($token) == ""} {
						array unset _registeredHandlers $token;
					};
				}
			} elseif {[_exists $widgetOrToken]} {
				foreach widget $_registeredHandlers($token) {
					_unregister $widget;
				}

				return;
			}

			array unset _registeredHandlers $widgetOrToken;

			return;
		}

		proc ~[this] {} {
			_off;

			_reset all;

			rename [this]	"";
			namespace delete [this];

			package forget [this];

			return;
		}

		proc [this] {} {
			variable _state;
			variable _index;
			variable _level;
			variable _caller;
			variable _namespace;
			variable _handlers;
			variable _registeredHandlers;

			set _state						0;
			set _index						1;
			set _level						0;
			set _caller						{};
			set _namespace					"";
			set _handlers()				{};
			set _registeredHandlers()	{};

			utils::buildOptions;

			_on;

			return;
		}

		[this]
	}

	unset {};
}

